/* $Id: 546x.c,v 1.9 1998/11/24 18:33:31 marcus Exp $
******************************************************************************

   Cirrus Logic CL-GD546x acceleration driver

   Copyright (C) 1997 Rodolphe Ortalo	[ortalo@laas.fr]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER      Rodolphe_Ortalo
#define DRIVER_NAME	"Cirrus Logic CL-GD546x (Laguna) Graphics Acceleration"
#define	DRIVER_VERSION	"$Revision: 1.9 $"

/*
	Kernel includes
*/
#define	request_io_region	request_region
#define	check_io_region		check_region  

/*
     KGI includes
*/

#include <kgi/module.h>
#include <kgi/kgi_commands.h>

#include "Cirrus/546x.h"

/* Ugly ? -- ortalo */
extern int CLIO_in8(int);
extern void CLIO_out8(int,int);
extern int CLIO_in16(int);
extern void CLIO_out16(int,int);
extern int CLIO_in32(int);
extern void CLIO_out32(int,int);
extern struct mem_region linFrameBuffer;

#include "suggest_546x.inc"

/*
**	Functions calling the 2D hardware accelerator
*/
/* Extracted from XFree86 source code (blame them then :-)
     (NB: STATUS is said read only in the manual....) */
static void reset_blt_engine(void)
{
	int ctrl = 0;
	DEBUG2("RESET0 Status=%x Qfree=%x Timeout=%x",CLIO_in16(D2_STATUS),
	       CLIO_in8(D2_QFREE), CLIO_in8(D2_TIMEOUT));
	CLIO_out16(D2_STATUS, CLIO_in16(D2_STATUS) ^ 0x04);
	CLIO_out16(D2_STATUS, CLIO_in16(D2_STATUS) ^ 0x04);
	/* flush the fifo */
	ctrl = CLIO_in16(D2_CONTROL);
	CLIO_out16(D2_CONTROL, ctrl | D2_FLUSH_WRITE_FIFO);
	CLIO_out16(D2_CONTROL, ctrl);
	DEBUG2("RESETn Status=%x Qfree=%x Timeout=%x",CLIO_in16(D2_STATUS),
	       CLIO_in8(D2_QFREE), CLIO_in8(D2_TIMEOUT));
}
static inline int wait_blt_ready(void)
{
	int c = 10000;
	while ( (CLIO_in16(D2_STATUS) & 0x7 /*D2_STATUS_BLT_READY*/)
		&& --c )
	{
		schedule();
	};
	if (!c)		reset_blt_engine();
	return (!c);
}
static inline int do_colorfill_bitblt(int x, int y, int w, int h, int bgcolor)
{
	if (wait_blt_ready())
	{
		NOTICE("BitBLT engine locked up before access !");
		return -E(ACCEL,NOSUP_NOW_MMAP);
	}
	CLIO_out32(D2_OP_BGCOLOR, bgcolor);
	CLIO_out16(D2_BLTDEF, D2_RES_IN_RDRAM | D2_OP0_IN_RDRAM
			| D2_OP1_COLORFILL | D2_OP2_COLORFILL);
	CLIO_out16(D2_DRAWDEF, D2_OP1);
	CLIO_out32(D2_OP0_RDRAM, ( x & D2_X)
			| ( (y << 16) & D2_Y ) );
	CLIO_out32(D2_BLTEXT_EX, ( w & D2_X_EXTENT)
			| ( (h << 16) & D2_Y_EXTENT ) );
	if (wait_blt_ready())
	{
		NOTICE("BitBLT engine locked up after access !");
		return -E(ACCEL,NOSUP_NOW_MMAP);
	}
	return EOK;
}
/*
**	Hardware accelerated drawing functions
**	provided by the driver.
*/
/*
**	IOCTL interface
*/
/* 2D functions
*/
static inline int cl546x_drawbox(struct kgi_graphic *gr, void *iobuf)
{	
	/* provides x,y,w,h */
	struct kgi_box *boxp = (struct kgi_box *)iobuf;

	if (gr->dev.mode.graphtype == KGIGT_4BIT || 
	    gr->dev.mode.graphtype == KGIGT_TEXT16)
	{
		return -E(ACCEL,NOSUP_MODE_MMAP);
	}
	if (boxp->w && boxp->h) {	/* 0 width is not OK! */
		return do_colorfill_bitblt(boxp->x, boxp->y, boxp->w, boxp->h,
			gr->gc->fg_color);
	}
	return EOK;
}

/*
**	KGI module interface
*/

int kgim_accel_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " " DRIVER_VERSION " loaded");
	/* !!! Check for Chipset and Version here !!! */
	return EOK;
}

void kgim_accel_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed");
}

int kgim_accel_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd *cmd)
{
	/* !!! We should probably check for accel-friendly dimensions here */
	return EOK;
}

void kgim_accel_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct chipset_private *chipsetprivate = KGIM_PRIVATE(dpy,chipset);
	kgi_uint n_tiles = chipsetprivate->tiles_per_line;
	kgi_uint wide_tile = chipsetprivate->wide_tiles;
	int ctrl_bpp = 0;
	int lncntl_bpp = 0;

	DEBUG1("From chipset private: n_tiles=%d, wide=%d", n_tiles, wide_tile);

	if (mode->request.graphtype != KGIGT_TEXT16) {
	  switch (mode->request.graphtype) {
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		case KGIGT_TEXT16:
		case KGIGT_TEXT32:
			NOTICE1("No acceleration in this mode: skipping init");
			return;
			break;
		case KGIGT_8BIT:
			ctrl_bpp = D2_8BPP;
			lncntl_bpp = D2_G_FORMAT_8BPP;
			break;
		case KGIGT_15BIT: /* Check this !!*/
			ctrl_bpp = D2_16BPP; 
			lncntl_bpp = D2_G_FORMAT_15BPP;
			break;
		case KGIGT_16BIT:
			ctrl_bpp = D2_16BPP; 
			lncntl_bpp = D2_G_FORMAT_16BPP;
			break;
		case KGIGT_24BIT:
			ctrl_bpp = D2_24BPP; 
			lncntl_bpp = D2_G_FORMAT_24BPP;
			break;
		case KGIGT_32BIT:
			ctrl_bpp = D2_32BPP; 
			lncntl_bpp = D2_G_FORMAT_32BPP;
			break;
		default:
			INTERNAL_ERROR;
			return;
		}
		/*
		**	Reset the Accelerator Engine
		*/
		NOTICE2("Resetting 2D accelerator");
		if (wait_blt_ready())
		{
			NOTICE("BitBLT engine locked up on init start !");
		}
		/* control register */
	   /* Should we be more careful with SWIZ bit ? (pp.8-18) -- ortalo */
		CLIO_out16(D2_CONTROL,  ctrl_bpp | D2_TILING_ENABLE
			| (wide_tile ? D2_WIDE_TILES : D2_SMALL_TILES) );
/* WARNING (wait?):		CLIO_out16(D2_LNCNTL, lncntl_bpp);*/
		/* Tile control */
		CLIO_out8(D2_TILE_CTRL, D2_TWO_WAY_INTERLEAVE 
			| ( n_tiles & D2_TILES_PER_LINE_MASK ) );
		/* Don't know how to use this:
		*/
		CLIO_out16(D2_TAG_MASK, 0x00);
		/* Disable bus timeout on fifo full */
		CLIO_out8(D2_TIMEOUT, D2_TIMEOUT_DISABLE);
/*		CLIO_out8(D2_TIMEOUT, D2_MAX_TIMEOUT); */

		/* No offset, framebuffer starts at 0,0 */
		CLIO_out8(D2_OFFSET_2D, 0x00);
		/* Reset bit mask */
		CLIO_out32(D2_BITMASK, 0xFFFFFFFF);
		if (wait_blt_ready())
		{
			NOTICE("BitBLT engine locked up on init end!");
		}
	}
	return;
}

int kgim_accel_command (struct kgi_graphic *gr, kgi_uint cmd,kgi_uint arg)
{ 
	kgi_suggest *sugg;
	const char *name;

	if (! (gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(ACCEL,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen ! */
	}
	switch(cmd) { 
		case ACCEL_GETSUGGEST:
			NOTICE2("Doing ACCEL_GETSUGGEST");
			sugg=(kgi_suggest *)(gr->io_buf);
			gr->io_res=gr->io_buf;

			if (sugg->id<0 || sugg->id > 4)
				return -E(ACCEL,BOGUS);
			name=sugg_default[gr->dev.mode.graphtype].text[sugg->id];
			if (!name)
				return -E(ACCEL,BOGUS);
			strcpy(sugg->name,name);
			sugg->args[0]=0;
			sugg->id++;	/* Next... */
			if (sugg->id >= 4 || sugg_default[gr->dev.mode.graphtype].text[sugg->id]==NULL )
			  sugg->id=0;
			return 0;

		/* Really accelerated functions. */

		case ACCEL_DRAWBOX:
			return(cl546x_drawbox(gr, gr->io_buf));
		
		case ACCEL_COPYBOX:
			return -E(ACCEL,NOSUP_ALWAYS_MMAP);
/*			return(s3_copybox(gr, gr->io_buf));
*/		
		case ACCEL_DRAWLINE:
			return -E(ACCEL,NOSUP_ALWAYS_MMAP);
/*			return(s3_drawline(gr, gr->io_buf));
*/
			/* Functions that will probably benefit 
			** from lower-lever accel 
			*/
		
		case ACCEL_FILLSCREEN:
			return -E(ACCEL,NOSUP_ALWAYS_MMAP);

		/* Functions that should probably use MMAP
		*/
		default: 
			return -E(ACCEL,NOSUP_ALWAYS_MMAP);

/* These functions cannot be supported via standard interfaces due to
** pointers to userspace in the args, but are efficiently emulated via MMAP:
** ACCEL_PUTHLINE	ACCEL_GETHLINE	ACCEL_PUTVLINE
** ACCEL_GETVLINE	 ACCEL_GETBOX	ACCEL_PUTBOX
*/

/* These are not implemented due to efficient support via MMAP.
** ACCEL_PUTPIXEL	ACCEL_GETPIXEL	ACCEL_DRAWPIXEL
** ACCEL_DRAWCIRCLE	ACCEL_DRAWHLINE	ACCEL_DRAWVLINE
*/
	}
	return -E(ACCEL,NOSUP_ALWAYS_CANT);/* This driver can't do this. */
}


